import dayjs from "dayjs"
import {LocationQueryValue} from "vue-router"
import {QueryConverter} from "@/components/UiScrollListModel"
import {uiComponentConfig} from "@/config/uiComponentConfig"

// UiPeriodSearch 내부에서 사용하는 date format
export const PERIOD_SEARCH_DATE_FORMAT = 'YYYYMMDD'

export type DatePeriod = {
  fromDate?: Date
  toDate?: Date
}

const serializeDate = (date?: Date, format = 'YYYYMMDD'): string => {
  return date ? dayjs(date).format(format) : ''
}

const deserializeDate = (str: string, endOfDate = false, format = 'YYYYMMDD'): Date | undefined => {
  if (str) {
    const date = dayjs(str, format)
    return endOfDate ? date.endOf('date').toDate() : date.toDate()
  } else {
    return undefined
  }
}

/**
 * DatePeriod 객체에 toString, toQueryValue, fromQueryValue 함수를 설정한다.
 * @param period
 */
export const enhanceDatePeriodObject = (period: DatePeriod): void => {

  period.toString = (): string => `${serializeDate(period.fromDate) || ''}_${serializeDate(period.toDate)}`

}

export const datePeriodConverter: QueryConverter<DatePeriod> = {

  toQueryString(value?: DatePeriod): string {
    const period = value as DatePeriod
    return `${serializeDate(period?.fromDate)}_${serializeDate(period?.toDate)}`
  },

  fromQueryString(query: LocationQueryValue | LocationQueryValue[]): DatePeriod | undefined {
    // 첫 번째 항목만 사용
    const str = String(Array.isArray(query) ? query[0] : query)
    if (str.includes('_')) {
      const dates = str.split('_')
      return {
        fromDate: deserializeDate(dates[0]),
        toDate: deserializeDate(dates[1], true)
      }
    }
    return undefined
  }
}

/**
 * months 로부터 DatePeriod 객체를 생성한다.
 * 예를 들어 months=3 이면 3개월 전부터 현재까지의 DatePeriod 가 생성된다.
 */
export const createDatePeriodByMonths = (months: number): DatePeriod => {
  const fromDate = dayjs().subtract(months, 'month').startOf('date').toDate()
  const toDate = dayjs().endOf('date').toDate()
  return {fromDate, toDate}
}

/**
 * yearDelta 로부터 DatePeriod 객체를 생성한다.
 * 예를 들어 yearDelta=-1 이면, 작년 1월 1일 ~ 작년 12월 31일 까지의 DatePeriod 가 생성된다.
 * yearDelta=0 이면 올해 1월 1일부터 오늘까지의 DatePeriod 가 생성된다.
 */
export const createDatePeriodByYearDelta = (yearDelta: number): DatePeriod => {
  const fromDate = dayjs().add(yearDelta, 'year').startOf('year').toDate()
  const toDate = ((yearDelta === 0) ? dayjs().endOf('date') : dayjs().add(yearDelta, 'year').endOf('year')).toDate()
  return {fromDate, toDate}
}

const evaluateCaption = (caption: string, period?: DatePeriod): string => {
  const from = period ? dayjs(period.fromDate) : undefined
  const to = period ? dayjs(period.toDate) : undefined
  return caption
      .replaceAll('{FROM_YEAR}', from?.format('YYYY') || '')
      .replaceAll('{FROM_MONTH}', from?.format('MM') || '')
      .replaceAll('{TO_YEAR}', to?.format('YYYY') || '')
      .replaceAll('{TO_MONTH}', to?.format('MM') || '')
}

export const normalizedPeriodSearchButton = (button: PeriodSearchButton): PeriodSearchButton => {
  if (!button._periodValue || !button._formattedValue) {
    const period = button.months
        ? createDatePeriodByMonths(button.months)
        : (button.yearDelta || button.yearDelta === 0)
            ? createDatePeriodByYearDelta(button.yearDelta)
            : undefined
    if (period) {
      button._periodValue = period
      button._formattedValue = {
        fromDate: dayjs(period.fromDate).format(PERIOD_SEARCH_DATE_FORMAT),
        toDate: dayjs(period.toDate).format(PERIOD_SEARCH_DATE_FORMAT)
      }
    }
  }
  button._evaluatedCaption = evaluateCaption(button.caption, button._periodValue)
  return button
}

/**
 * UiDatePeriod 컴포넌트의 초기 값을 가져온다.
 */
export const getPeriodSearchInitialValue = (button?: PeriodSearchButton): DatePeriod | undefined => {
  const firstButton = button || uiComponentConfig.periodSearchButtons?.[0]
  return normalizedPeriodSearchButton(firstButton)._periodValue
}

export type PeriodSearchButton = {
  caption: string
  months?: number
  yearDelta?: number
  _periodValue?: DatePeriod // internal use only
  _formattedValue?: { fromDate: string, toDate: string } // internal use only
  _evaluatedCaption?: string // internal use only
}

/**
 * 두 DatePeriod 객체가 동일한지 확인한다.
 * @param a
 * @param b
 */
export const isDatePeriodEqual = (a?: DatePeriod, b?: DatePeriod): boolean => {
  if (a === undefined && b === undefined) return true
  else if (a === undefined || b === undefined) return false
  else return JSON.stringify(a, Object.keys(a).sort()) === JSON.stringify(b, Object.keys(b).sort())
}